import sempy.fabric as fabric
import re
import json
import base64
import time
import uuid
from sempy.fabric.exceptions import FabricHTTPException
import pandas as pd
from functools import wraps
import datetime
from typing import Optional, Tuple, List
from uuid import UUID
import sempy_labs._icons as icons
import urllib.parse
from azure.core.credentials import TokenCredential, AccessToken


def create_abfss_path(
    lakehouse_id: UUID, lakehouse_workspace_id: UUID, delta_table_name: str
) -> str:
    """
    Creates an abfss path for a delta table in a Fabric lakehouse.

    Parameters
    ----------
    lakehouse_id : UUID
        ID of the Fabric lakehouse.
    lakehouse_workspace_id : UUID
        ID of the Fabric workspace.
    delta_table_name : str
        Name of the delta table name.

    Returns
    -------
    str
        An abfss path which can be used to save/reference a delta table in a Fabric lakehouse.
    """

    return f"abfss://{lakehouse_workspace_id}@onelake.dfs.fabric.microsoft.com/{lakehouse_id}/Tables/{delta_table_name}"


def format_dax_object_name(table: str, column: str) -> str:
    """
    Formats a table/column combination to the 'Table Name'[Column Name] format.

    Parameters
    ----------
    table : str
        The name of the table.
    column : str
        The name of the column.

    Returns
    -------
    str
        The fully qualified object name.
    """

    return "'" + table + "'[" + column + "]"


def create_relationship_name(
    from_table: str, from_column: str, to_table: str, to_column: str
) -> str:
    """
    Formats a relationship's table/columns into a fully qualified name.

    Parameters
    ----------
    from_table : str
        The name of the table on the 'from' side of the relationship.
    from_column : str
        The name of the column on the 'from' side of the relationship.
    to_table : str
        The name of the table on the 'to' side of the relationship.
    to_column : str
        The name of the column on the 'to' side of the relationship.

    Returns
    -------
    str
        The fully qualified relationship name.
    """

    return (
        format_dax_object_name(from_table, from_column)
        + " -> "
        + format_dax_object_name(to_table, to_column)
    )


def resolve_report_id(report: str, workspace: Optional[str] = None) -> UUID:
    """
    Obtains the ID of the Power BI report.

    Parameters
    ----------
    report : str
        The name of the Power BI report.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    UUID
        The ID of the Power BI report.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    obj = fabric.resolve_item_id(item_name=report, type="Report", workspace=workspace)

    return obj


def resolve_report_name(report_id: UUID, workspace: Optional[str] = None) -> str:
    """
    Obtains the name of the Power BI report.

    Parameters
    ----------
    report_id : UUID
        The name of the Power BI report.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The name of the Power BI report.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    obj = fabric.resolve_item_name(
        item_id=report_id, type="Report", workspace=workspace
    )

    return obj


def resolve_dataset_id(dataset: str, workspace: Optional[str] = None) -> UUID:
    """
    Obtains the ID of the semantic model.

    Parameters
    ----------
    dataset : str
        The name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    UUID
        The ID of the semantic model.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    obj = fabric.resolve_item_id(
        item_name=dataset, type="SemanticModel", workspace=workspace
    )

    return obj


def resolve_dataset_name(dataset_id: UUID, workspace: Optional[str] = None) -> str:
    """
    Obtains the name of the semantic model.

    Parameters
    ----------
    dataset_id : UUID
        The name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The name of the semantic model.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    obj = fabric.resolve_item_name(
        item_id=dataset_id, type="SemanticModel", workspace=workspace
    )

    return obj


def resolve_lakehouse_name(
    lakehouse_id: Optional[UUID] = None, workspace: Optional[str] = None
) -> str:
    """
    Obtains the name of the Fabric lakehouse.

    Parameters
    ----------
    lakehouse_id : UUID, default=None
        The name of the Fabric lakehouse.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The name of the Fabric lakehouse.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    if lakehouse_id is None:
        lakehouse_id = fabric.get_lakehouse_id()

    obj = fabric.resolve_item_name(
        item_id=lakehouse_id, type="Lakehouse", workspace=workspace
    )

    return obj


def resolve_lakehouse_id(lakehouse: str, workspace: Optional[str] = None) -> UUID:
    """
    Obtains the ID of the Fabric lakehouse.

    Parameters
    ----------
    lakehouse : str
        The name of the Fabric lakehouse.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    UUID
        The ID of the Fabric lakehouse.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    obj = fabric.resolve_item_id(
        item_name=lakehouse, type="Lakehouse", workspace=workspace
    )

    return obj


def get_direct_lake_sql_endpoint(dataset: str, workspace: Optional[str] = None) -> UUID:
    """
    Obtains the SQL Endpoint ID of the semantic model.

    Parameters
    ----------
    dataset : str
        The name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The ID of SQL Endpoint.
    """

    from sempy_labs.tom import connect_semantic_model

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)

    # dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    # dfP_filt = dfP[dfP["Mode"] == "DirectLake"]

    # if len(dfP_filt) == 0:
    #    raise ValueError(
    #        f"The '{dataset}' semantic model in the '{workspace}' workspace is not in Direct Lake mode."
    #    )

    with connect_semantic_model(
        dataset=dataset, readonly=True, workspace=workspace
    ) as tom:
        sqlEndpointId = None
        for e in tom.model.Expressions:
            if e.Name == "DatabaseQuery":
                expr = e.Expression
                matches = re.findall(r'"([^"]+)"', expr)
                sqlEndpointId = matches[1]

        if sqlEndpointId is None:
            raise ValueError("SQL Endpoint not found.")

        return sqlEndpointId


def generate_embedded_filter(filter: str) -> str:
    """
    Converts the filter expression to a filter expression which can be used by a Power BI embedded URL.

    Parameters
    ----------
    filter : str
        The filter expression for an embedded Power BI report.

    Returns
    -------
    str
        A filter expression usable by a Power BI embedded URL.
    """

    pattern = r"'[^']+'\[[^\[]+\]"
    matches = re.findall(pattern, filter)
    for match in matches:
        matchReplace = (
            match.replace("'", "")
            .replace("[", "/")
            .replace("]", "")
            .replace(" ", "_x0020_")
            .replace("@", "_00x40_")
            .replace("+", "_0x2B_")
            .replace("{", "_007B_")
            .replace("}", "_007D_")
        )
        filter = filter.replace(match, matchReplace)

    pattern = r"\[[^\[]+\]"
    matches = re.findall(pattern, filter)
    for match in matches:
        matchReplace = (
            match.replace("'", "")
            .replace("[", "/")
            .replace("]", "")
            .replace(" ", "_x0020_")
            .replace("@", "_00x40_")
            .replace("+", "_0x2B_")
            .replace("{", "_007B_")
            .replace("}", "_007D_")
        )
        filter = filter.replace(match, matchReplace)

    revised_filter = (
        filter.replace("<=", "le")
        .replace(">=", "ge")
        .replace("<>", "ne")
        .replace("!=", "ne")
        .replace("==", "eq")
        .replace("=", "eq")
        .replace("<", "lt")
        .replace(">", "gt")
        .replace(" && ", " and ")
        .replace(" & ", " and ")
        .replace(" || ", " or ")
        .replace(" | ", " or ")
        .replace("{", "(")
        .replace("}", ")")
    )

    return revised_filter


def save_as_delta_table(
    dataframe,
    delta_table_name: str,
    write_mode: str,
    merge_schema: bool = False,
    schema: Optional[dict] = None,
    lakehouse: Optional[str] = None,
    workspace: Optional[str] = None,
):
    """
    Saves a pandas dataframe as a delta table in a Fabric lakehouse.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The dataframe to be saved as a delta table.
    delta_table_name : str
        The name of the delta table.
    write_mode : str
        The write mode for the save operation. Options: 'append', 'overwrite'.
    merge_schema : bool, default=False
        Merges the schemas of the dataframe to the delta table.
    schema : dict, default=None
        A dictionary showing the schema of the columns and their data types.
    lakehouse : str, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    from pyspark.sql import SparkSession
    from pyspark.sql.types import (
        StringType,
        IntegerType,
        FloatType,
        DateType,
        StructType,
        StructField,
        BooleanType,
        LongType,
        DoubleType,
        TimestampType,
    )

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    if lakehouse is None:
        lakehouse_id = fabric.get_lakehouse_id()
        lakehouse = resolve_lakehouse_name(
            lakehouse_id=lakehouse_id, workspace=workspace
        )
    else:
        lakehouse_id = resolve_lakehouse_id(lakehouse, workspace)

    writeModes = ["append", "overwrite"]
    write_mode = write_mode.lower()

    if write_mode not in writeModes:
        raise ValueError(
            f"{icons.red_dot} Invalid 'write_type' parameter. Choose from one of the following values: {writeModes}."
        )

    if " " in delta_table_name:
        raise ValueError(
            f"{icons.red_dot} Invalid 'delta_table_name'. Delta tables in the lakehouse cannot have spaces in their names."
        )

    dataframe.columns = dataframe.columns.str.replace(" ", "_")
    spark = SparkSession.builder.getOrCreate()

    type_mapping = {
        "string": StringType(),
        "str": StringType(),
        "integer": IntegerType(),
        "int": IntegerType(),
        "float": FloatType(),
        "date": DateType(),
        "bool": BooleanType(),
        "boolean": BooleanType(),
        "long": LongType(),
        "double": DoubleType(),
        "timestamp": TimestampType(),
    }

    if schema is None:
        spark_df = spark.createDataFrame(dataframe)
    else:
        schema_map = StructType(
            [
                StructField(column_name, type_mapping[data_type], True)
                for column_name, data_type in schema.items()
            ]
        )
        spark_df = spark.createDataFrame(dataframe, schema_map)

    filePath = create_abfss_path(
        lakehouse_id=lakehouse_id,
        lakehouse_workspace_id=workspace_id,
        delta_table_name=delta_table_name,
    )

    if merge_schema:
        spark_df.write.mode(write_mode).format("delta").option(
            "mergeSchema", "true"
        ).save(filePath)
    else:
        spark_df.write.mode(write_mode).format("delta").save(filePath)
    print(
        f"{icons.green_dot} The dataframe has been saved as the '{delta_table_name}' table in the '{lakehouse}' lakehouse within the '{workspace}' workspace."
    )


def language_validate(language: str):
    """
    Validateds that the language specified exists within the supported langauges.

    Parameters
    ----------
    language : str
        The language code.

    Returns
    -------
    bool
        A True/False indication as to whether the language code is supported.
    """

    url = "https://learn.microsoft.com/azure/ai-services/translator/language-support"

    tables = pd.read_html(url)
    df = tables[0]

    df_filt = df[df["Language code"] == language]

    df_filt2 = df[df["Language"] == language.capitalize()]

    if len(df_filt) == 1:
        lang = df_filt["Language"].iloc[0]
    elif len(df_filt2) == 1:
        lang = df_filt2["Language"].iloc[0]
    else:
        raise ValueError(
            f"{icons.red_dot} The '{language}' language is not a valid language code. Please refer to this link for a list of valid language codes: {url}."
        )

    return lang


def resolve_workspace_name_and_id(workspace: Optional[str] = None) -> Tuple[str, str]:
    """
    Obtains the name and ID of the Fabric workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str, str
        The name and ID of the Fabric workspace.
    """

    if workspace is None:
        workspace_id = fabric.get_workspace_id()
        workspace = fabric.resolve_workspace_name(workspace_id)
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    return str(workspace), str(workspace_id)


def _extract_json(dataframe: pd.DataFrame) -> dict:

    payload = dataframe["payload"].iloc[0]
    json_file = _decode_b64(payload)

    return json.loads(json_file)


def _conv_b64(file):

    loadJson = json.dumps(file)
    f = base64.b64encode(loadJson.encode("utf-8")).decode("utf-8")

    return f


def _decode_b64(file, format: Optional[str] = "utf-8"):

    result = base64.b64decode(file).decode(format)

    return result


def is_default_semantic_model(dataset: str, workspace: Optional[str] = None) -> bool:
    """
    Identifies whether a semantic model is a default semantic model.

    Parameters
    ----------
    dataset : str
        The name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    bool
        A True/False value indicating whether the semantic model is a default semantic model.
    """

    workspace = fabric.resolve_workspace_name(workspace)

    dfI = fabric.list_items(workspace=workspace)
    filtered_df = dfI.groupby("Display Name").filter(
        lambda x: set(["Warehouse", "SemanticModel"]).issubset(set(x["Type"]))
        or set(["Lakehouse", "SemanticModel"]).issubset(set(x["Type"]))
    )
    default_semantic_models = filtered_df["Display Name"].unique().tolist()

    return dataset in default_semantic_models


def resolve_item_type(item_id: UUID, workspace: Optional[str] = None) -> str:
    """
    Obtains the item type for a given Fabric Item Id within a Fabric workspace.

    Parameters
    ----------
    item_id : UUID
        The item/artifact Id.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The item type for the item Id.
    """

    workspace = fabric.resolve_workspace_name(workspace)
    dfI = fabric.list_items(workspace=workspace)
    dfI_filt = dfI[dfI["Id"] == item_id]

    if len(dfI_filt) == 0:
        raise ValueError(
            f"Invalid 'item_id' parameter. The '{item_id}' item was not found in the '{workspace}' workspace."
        )
    item_type = dfI_filt["Type"].iloc[0]

    return item_type


def resolve_dataset_from_report(
    report: str, workspace: Optional[str] = None
) -> Tuple[UUID, str, UUID, str]:
    """
    Obtains the basic semantic model properties from which the report's data is sourced.

    Parameters
    ----------
    report : str
        The name of the Power BI report.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[UUID, str, UUID, str]
        The semantic model UUID, semantic model name, semantic model workspace UUID, semantic model workspace name
    """

    workspace = fabric.resolve_workspace_name(workspace)

    dfR = fabric.list_reports(workspace=workspace)
    dfR_filt = dfR[dfR["Name"] == report]
    if len(dfR_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{report}' report does not exist within the '{workspace}' workspace."
        )
    dataset_id = dfR_filt["Dataset Id"].iloc[0]
    dataset_workspace_id = dfR_filt["Dataset Workspace Id"].iloc[0]
    dataset_workspace = fabric.resolve_workspace_name(dataset_workspace_id)
    dataset_name = resolve_dataset_name(
        dataset_id=dataset_id, workspace=dataset_workspace
    )

    return dataset_id, dataset_name, dataset_workspace_id, dataset_workspace


def _add_part(target_dict, path, payload):

    part = {"path": path, "payload": payload, "payloadType": "InlineBase64"}

    target_dict["definition"]["parts"].append(part)


def resolve_workspace_capacity(workspace: Optional[str] = None) -> Tuple[UUID, str]:
    """
    Obtains the capacity Id and capacity name for a given workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[UUID, str]
        capacity Id; capacity came.
    """

    workspace = fabric.resolve_workspace_name(workspace)
    filter_condition = urllib.parse.quote(workspace)
    dfW = fabric.list_workspaces(filter=f"name eq '{filter_condition}'")
    capacity_id = dfW["Capacity Id"].iloc[0]
    dfC = fabric.list_capacities()
    dfC_filt = dfC[dfC["Id"] == capacity_id]
    if len(dfC_filt) == 1:
        capacity_name = dfC_filt["Display Name"].iloc[0]
    else:
        capacity_name = None

    return capacity_id, capacity_name


def get_capacity_id(workspace: Optional[str] = None) -> UUID:
    """
    Obtains the Capacity Id for a given workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    UUID
        The capacity Id.
    """

    workspace = fabric.resolve_workspace_name(workspace)
    filter_condition = urllib.parse.quote(workspace)
    dfW = fabric.list_workspaces(filter=f"name eq '{filter_condition}'")
    if len(dfW) == 0:
        raise ValueError(f"{icons.red_dot} The '{workspace}' does not exist'.")

    return dfW["Capacity Id"].iloc[0]


def get_capacity_name(workspace: Optional[str] = None) -> str:
    """
    Obtains the capacity name for a given workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The capacity name.
    """

    capacity_id = get_capacity_id(workspace)
    dfC = fabric.list_capacities()
    dfC_filt = dfC[dfC["Id"] == capacity_id]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{capacity_id}' capacity Id does not exist."
        )

    return dfC_filt["Display Name"].iloc[0]


def resolve_capacity_name(capacity_id: Optional[UUID] = None) -> str:
    """
    Obtains the capacity name for a given capacity Id.

    Parameters
    ----------
    capacity_id : UUID, default=None
        The capacity Id.
        Defaults to None which resolves to the capacity name of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.

    Returns
    -------
    str
        The capacity name.
    """

    if capacity_id is None:
        return get_capacity_name()

    dfC = fabric.list_capacities()
    dfC_filt = dfC[dfC["Id"] == capacity_id]

    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{capacity_id}' capacity Id does not exist."
        )

    return dfC_filt["Display Name"].iloc[0]


def resolve_capacity_id(capacity_name: Optional[str] = None) -> UUID:
    """
    Obtains the capacity Id for a given capacity name.

    Parameters
    ----------
    capacity_name : str, default=None
        The capacity name.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.

    Returns
    -------
    UUID
        The capacity Id.
    """

    if capacity_name is None:
        return get_capacity_id()

    dfC = fabric.list_capacities()
    dfC_filt = dfC[dfC["Display Name"] == capacity_name]

    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{capacity_name}' capacity does not exist."
        )

    return dfC_filt["Id"].iloc[0]


def retry(sleep_time: int, timeout_error_message: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = datetime.datetime.now()
            timeout = datetime.timedelta(minutes=1)
            while datetime.datetime.now() - start_time <= timeout:
                try:
                    return func(*args, **kwargs)
                except Exception:
                    time.sleep(sleep_time)
            raise TimeoutError(timeout_error_message)

        return wrapper

    return decorator


def lro(
    client,
    response,
    status_codes: Optional[List[str]] = [200, 202],
    sleep_time: Optional[int] = 1,
    return_status_code: bool = False,
):

    if response.status_code not in status_codes:
        raise FabricHTTPException(response)
    if response.status_code == status_codes[0]:
        if return_status_code:
            result = response.status_code
        else:
            result = response
    if response.status_code == status_codes[1]:
        operationId = response.headers["x-ms-operation-id"]
        response = client.get(f"/v1/operations/{operationId}")
        response_body = json.loads(response.content)
        while response_body["status"] not in ["Succeeded", "Failed"]:
            time.sleep(sleep_time)
            response = client.get(f"/v1/operations/{operationId}")
            response_body = json.loads(response.content)
        if response_body["status"] != "Succeeded":
            raise FabricHTTPException(response)
        if return_status_code:
            result = response.status_code
        else:
            response = client.get(f"/v1/operations/{operationId}/result")
            result = response

    return result


def pagination(client, response):

    responses = []
    response_json = response.json()
    responses.append(response_json)

    # Check for pagination
    continuation_token = response_json.get("continuationToken")
    continuation_uri = response_json.get("continuationUri")

    # Loop to handle pagination
    while continuation_token is not None:
        response = client.get(continuation_uri)
        response_json = response.json()
        responses.append(response_json)

        # Update the continuation token and URI for the next iteration
        continuation_token = response_json.get("continuationToken")
        continuation_uri = response_json.get("continuationUri")

    return responses


def resolve_deployment_pipeline_id(deployment_pipeline: str) -> UUID:
    """
    Obtains the Id for a given deployment pipeline.

    Parameters
    ----------
    deployment_pipeline : str
        The deployment pipeline name

    Returns
    -------
    UUID
        The deployment pipeline Id.
    """

    from sempy_labs._deployment_pipelines import list_deployment_pipelines

    dfP = list_deployment_pipelines()
    dfP_filt = dfP[dfP["Deployment Pipeline Name"] == deployment_pipeline]
    if len(dfP_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{deployment_pipeline}' deployment pipeline is not valid."
        )
    deployment_pipeline_id = dfP_filt["Deployment Pipeline Id"].iloc[0]

    return deployment_pipeline_id


class FabricTokenCredential(TokenCredential):

    def get_token(
        self,
        scopes: str,
        claims: Optional[str] = None,
        tenant_id: Optional[str] = None,
        enable_cae: bool = False,
        **kwargs: any,
    ) -> AccessToken:

        import notebookutils

        token = notebookutils.credentials.getToken(scopes)
        access_token = AccessToken(token, 0)

        return access_token


def _get_adls_client(account_name):

    from azure.storage.filedatalake import DataLakeServiceClient

    account_url = f"https://{account_name}.dfs.core.windows.net"

    service_client = DataLakeServiceClient(
        account_url, credential=FabricTokenCredential()
    )

    return service_client


def resolve_warehouse_id(warehouse: str, workspace: Optional[str]) -> UUID:
    """
    Obtains the Id for a given warehouse.

    Parameters
    ----------
    warehouse : str
        The warehouse name

    Returns
    -------
    UUID
        The warehouse Id.
    """

    workspace = fabric.resolve_workspace_name(workspace)
    return fabric.resolve_item_id(
        item_name=warehouse, type="Warehouse", workspace=workspace
    )


def get_language_codes(languages: str | List[str]):

    if isinstance(languages, str):
        languages = [languages]

    for i, lang in enumerate(languages):
        for k, v in icons.language_map.items():
            if v == lang.capitalize():
                languages[i] = k
                break

    return languages


def _get_azure_token_credentials(
    key_vault_uri: str,
    key_vault_tenant_id: str,
    key_vault_client_id: str,
    key_vault_client_secret: str,
    audience: str = "https://management.azure.com/.default",
) -> Tuple[str, str, dict]:

    import notebookutils
    from azure.identity import ClientSecretCredential

    # "https://analysis.windows.net/powerbi/api/.default"

    tenant_id = notebookutils.credentials.getSecret(key_vault_uri, key_vault_tenant_id)
    client_id = notebookutils.credentials.getSecret(key_vault_uri, key_vault_client_id)
    client_secret = notebookutils.credentials.getSecret(
        key_vault_uri, key_vault_client_secret
    )

    credential = ClientSecretCredential(
        tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
    )

    token = credential.get_token(audience).token

    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }

    return token, credential, headers


def convert_to_alphanumeric_lowercase(input_string):

    # Removes non-alphanumeric characters
    cleaned_string = re.sub(r"[^a-zA-Z0-9]", "", input_string)
    cleaned_string = cleaned_string.lower()

    return cleaned_string


def resolve_environment_id(environment: str, workspace: Optional[str] = None) -> UUID:
    """
    Obtains the environment Id for a given environment.

    Parameters
    ----------
    environment: str
        Name of the environment.

    Returns
    -------
    UUID
        The environment Id.
    """

    workspace = fabric.resolve_workspace_name(workspace)
    return fabric.resolve_item_id(
        item_name=environment, type="Environment", workspace=workspace
    )


def _make_clickable(val):

    return f'<a target="_blank" href="{val}">{val}</a>'


def convert_to_friendly_case(text: str) -> str:
    """
    Converts a string of pascal/camel/snake case to business-friendly case.

    Parameters
    ----------
    text : str
        The text to convert.

    Returns
    -------
    str
        Text converted into a business-friendly text.
    """
    if text is not None:
        text = text.replace("_", " ")
        # Insert space before each capital letter, avoiding double spaces
        text = re.sub(r"(?<!\s)(?=[A-Z])", " ", text)
        # Strip leading/trailing whitespace and capitalize the first letter of each word
        text = text.strip().title()

    return text


def resolve_notebook_id(notebook: str, workspace: Optional[str] = None) -> UUID:
    """
    Obtains the notebook Id for a given notebook.

    Parameters
    ----------
    notebook: str
        Name of the notebook.

    Returns
    -------
    UUID
        The notebook Id.
    """

    workspace = fabric.resolve_workspace_name(workspace)
    return fabric.resolve_item_id(
        item_name=notebook, type="Notebook", workspace=workspace
    )


def generate_guid():

    return str(uuid.uuid4())


def _get_max_run_id(lakehouse: str, table_name: str) -> int:

    from pyspark.sql import SparkSession

    spark = SparkSession.builder.getOrCreate()
    query = f"SELECT MAX(RunId) FROM {lakehouse}.{table_name}"
    dfSpark = spark.sql(query)
    max_run_id = dfSpark.collect()[0][0]

    return max_run_id



import sempy.fabric as fabric
import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    lro,
    pagination,
)
from sempy.fabric.exceptions import FabricHTTPException


def list_ml_experiments(workspace: Optional[str] = None) -> pd.DataFrame:
    """
    Shows the ML experiments within a workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the ML models within a workspace.
    """

    df = pd.DataFrame(columns=["ML Experiment Name", "ML Experiment Id", "Description"])

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    client = fabric.FabricRestClient()
    response = client.get(f"/v1/workspaces/{workspace_id}/mlExperiments")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for v in r.get("value", []):
            model_id = v.get("id")
            modelName = v.get("displayName")
            desc = v.get("description")

            new_data = {
                "ML Experiment Name": modelName,
                "ML Experiment Id": model_id,
                "Description": desc,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def create_ml_experiment(
    name: str, description: Optional[str] = None, workspace: Optional[str] = None
):
    """
    Creates a Fabric ML experiment.

    Parameters
    ----------
    name: str
        Name of the ML experiment.
    description : str, default=None
        A description of the environment.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    request_body = {"displayName": name}

    if description:
        request_body["description"] = description

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/workspaces/{workspace_id}/mlExperiments", json=request_body
    )

    lro(client, response, status_codes=[201, 202])

    print(
        f"{icons.green_dot} The '{name}' ML experiment has been created within the '{workspace}' workspace."
    )


def delete_ml_experiment(name: str, workspace: Optional[str] = None):
    """
    Deletes a Fabric ML experiment.

    Parameters
    ----------
    name: str
        Name of the ML experiment.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    item_id = fabric.resolve_item_id(
        item_name=name, type="MLExperiment", workspace=workspace
    )

    client = fabric.FabricRestClient()
    response = client.delete(f"/v1/workspaces/{workspace_id}/mlExperiments/{item_id}")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    print(
        f"{icons.green_dot} The '{name}' ML experiment within the '{workspace}' workspace has been deleted."
    )



import sempy.fabric as fabric
from sempy_labs._helper_functions import (
    resolve_report_id,
    format_dax_object_name,
    resolve_dataset_from_report,
    _conv_b64,
    _extract_json,
    _add_part,
    lro,
    # _make_clickable,
)
from typing import Optional, List
import pandas as pd
import re
import json
import base64
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons
import sempy_labs.report._report_helper as helper
from sempy_labs._model_dependencies import get_measure_dependencies
from jsonpath_ng.ext import parse
import warnings


class ReportWrapper:

    _report: str
    _workspace: str

    @log
    def __init__(
        self,
        report: str,
        workspace: Optional[str] = None,
    ):

        from sempy_labs.report import get_report_definition

        warnings.simplefilter(action="ignore", category=FutureWarning)

        self._report = report
        self._workspace = workspace
        self._workspace_id = fabric.resolve_workspace_id(workspace)
        self._report_id = resolve_report_id(report, workspace)
        self.rdef = get_report_definition(
            report=self._report, workspace=self._workspace
        )

        if len(self.rdef[self.rdef["path"] == "definition/report.json"]) == 0:
            raise ValueError(
                f"{icons.red_dot} The ReportWrapper function requires the report to be in the PBIR format."
                "See here for details: https://powerbi.microsoft.com/blog/power-bi-enhanced-report-format-pbir-in-power-bi-desktop-developer-mode-preview/"
            )

    # Helper functions
    def _add_extended(self, dataframe):

        from sempy_labs.tom import connect_semantic_model

        dataset_id, dataset_name, dataset_workspace_id, dataset_workspace = (
            resolve_dataset_from_report(report=self._report, workspace=self._workspace)
        )

        with connect_semantic_model(
            dataset=dataset_name, readonly=True, workspace=dataset_workspace
        ) as tom:
            for index, row in dataframe.iterrows():
                obj_type = row["Object Type"]
                if obj_type == "Measure":
                    dataframe.at[index, "Valid Semantic Model Object"] = any(
                        o.Name == row["Object Name"] for o in tom.all_measures()
                    )
                elif obj_type == "Column":
                    dataframe.at[index, "Valid Semantic Model Object"] = any(
                        format_dax_object_name(c.Parent.Name, c.Name)
                        == format_dax_object_name(row["Table Name"], row["Object Name"])
                        for c in tom.all_columns()
                    )
                elif obj_type == "Hierarchy":
                    dataframe.at[index, "Valid Semantic Model Object"] = any(
                        format_dax_object_name(h.Parent.Name, h.Name)
                        == format_dax_object_name(row["Table Name"], row["Object Name"])
                        for h in tom.all_hierarchies()
                    )
        return dataframe

    def update_report(self, request_body: dict):

        client = fabric.FabricRestClient()
        response = client.post(
            f"/v1/workspaces/{self._workspace_id}/reports/{self._report_id}/updateDefinition",
            json=request_body,
        )

        lro(client, response, return_status_code=True)

    def resolve_page_name(self, page_display_name: str) -> UUID:
        """
        Obtains the page name, page display name, and the file path for a given page in a report.

        Parameters
        ----------
        page_display_name : str
            The display name of the page of the report.

        Returns
        -------
        UUID
            The page name.
        """

        x, y, z = helper.resolve_page_name(self, page_display_name)

        return x

    def resolve_page_display_name(self, page_name: UUID) -> str:
        """
        Obtains the page dispaly name.

        Parameters
        ----------
        page_name : UUID
            The name of the page of the report.

        Returns
        -------
        str
            The page display name.
        """

        x, y, z = helper.resolve_page_name(self, page_name=page_name)

        return y

    def get_theme(self, theme_type: str = "baseTheme") -> dict:
        """
        Obtains the theme file of the report.

        Parameters
        ----------
        theme_type : str, default="baseTheme"
            The theme type. Options: "baseTheme", "customTheme".

        Returns
        -------
        dict
            The theme.json file
        """

        theme_types = ["baseTheme", "customTheme"]
        theme_type = theme_type.lower()

        if "custom" in theme_type:
            theme_type = "customTheme"
        elif "base" in theme_type:
            theme_type = "baseTheme"
        if theme_type not in theme_types:
            raise ValueError(
                f"{icons.red_dot} Invalid theme type. Valid options: {theme_types}."
            )

        rptdef = self.rdef[self.rdef["path"] == "definition/report.json"]
        rptJson = _extract_json(rptdef)
        theme_collection = rptJson.get("themeCollection", {})
        if theme_type not in theme_collection:
            raise ValueError(
                f"{icons.red_dot} The {self._report} report within the '{self._workspace} workspace has no custom theme."
            )
        ct = theme_collection.get(theme_type)
        theme_name = ct["name"]
        theme_location = ct["type"]
        theme_file_path = f"StaticResources/{theme_location}/{theme_name}"
        if theme_type == "baseTheme":
            theme_file_path = (
                f"StaticResources/{theme_location}/BaseThemes/{theme_name}"
            )
        if not theme_file_path.endswith(".json"):
            theme_file_path = f"{theme_file_path}.json"

        theme_df = self.rdef[self.rdef["path"] == theme_file_path]
        theme_json = _extract_json(theme_df)

        return theme_json

    # List functions
    def list_custom_visuals(self) -> pd.DataFrame:
        """
        Shows a list of all custom visuals used in the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all the custom visuals used in the report.
        """

        helper.populate_custom_visual_display_names()

        df = pd.DataFrame(columns=["Custom Visual Name", "Custom Visual Display Name"])
        rd = self.rdef
        rd_filt = rd[rd["path"] == "definition/report.json"]
        rptJson = _extract_json(rd_filt)
        df["Custom Visual Name"] = rptJson.get("publicCustomVisuals")
        df["Custom Visual Display Name"] = df["Custom Visual Name"].apply(
            lambda x: helper.vis_type_mapping.get(x, x)
        )

        df["Used in Report"] = df["Custom Visual Name"].isin(
            self.list_visuals()["Type"]
        )

        bool_cols = ["Used in Report"]
        df[bool_cols] = df[bool_cols].astype(bool)

        return df

    def list_report_filters(self, extended: bool = False) -> pd.DataFrame:
        """
        Shows a list of all report filters used in the report.

        Parameters
        ----------
        extended : bool, default=False
            If True, adds an extra column called 'Valid Semantic Model Object' which identifies whether the semantic model object used
            in the report exists in the semantic model which feeds data to the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all the report filters used in the report.
        """

        rd_filt = self.rdef[self.rdef["path"] == "definition/report.json"]

        df = pd.DataFrame(
            columns=[
                "Filter Name",
                "Type",
                "Table Name",
                "Object Name",
                "Object Type",
                "Hidden",
                "Locked",
                "How Created",
                "Used",
            ]
        )

        if len(rd_filt) == 1:
            rpt_json = _extract_json(rd_filt)
            if "filterConfig" in rpt_json:
                for flt in rpt_json.get("filterConfig", {}).get("filters", {}):
                    filter_name = flt.get("name")
                    how_created = flt.get("howCreated")
                    locked = flt.get("isLockedInViewMode", False)
                    hidden = flt.get("isHiddenInViewMode", False)
                    filter_type = flt.get("type", "Basic")
                    filter_used = True if "Where" in flt.get("filter", {}) else False

                    entity_property_pairs = helper.find_entity_property_pairs(flt)

                    for object_name, properties in entity_property_pairs.items():
                        new_data = {
                            "Filter Name": filter_name,
                            "Type": filter_type,
                            "Table Name": properties[0],
                            "Object Name": object_name,
                            "Object Type": properties[1],
                            "Hidden": hidden,
                            "Locked": locked,
                            "How Created": how_created,
                            "Used": filter_used,
                        }

                        df = pd.concat(
                            [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                        )

        bool_cols = ["Hidden", "Locked", "Used"]
        df[bool_cols] = df[bool_cols].astype(bool)

        if extended:
            df = self._add_extended(dataframe=df)

        return df

    def list_page_filters(self, extended: bool = False) -> pd.DataFrame:
        """
        Shows a list of all page filters used in the report.

        Parameters
        ----------
        extended : bool, default=False
            If True, adds an extra column called 'Valid Semantic Model Object' which identifies whether the semantic model object used
            in the report exists in the semantic model which feeds data to the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all the page filters used in the report.
        """

        rd = self.rdef
        df = pd.DataFrame(
            columns=[
                "Page Name",
                "Page Display Name",
                "Filter Name",
                "Type",
                "Table Name",
                "Object Name",
                "Object Type",
                "Hidden",
                "Locked",
                "How Created",
                "Used",
            ]
        )

        for _, r in rd.iterrows():
            path = r["path"]
            payload = r["payload"]
            if path.endswith("/page.json"):
                obj_file = base64.b64decode(payload).decode("utf-8")
                obj_json = json.loads(obj_file)
                page_id = obj_json.get("name")
                page_display = obj_json.get("displayName")

                if "filterConfig" in obj_json:
                    for flt in obj_json.get("filterConfig", {}).get("filters", {}):
                        filter_name = flt.get("name")
                        how_created = flt.get("howCreated")
                        locked = flt.get("isLockedInViewMode", False)
                        hidden = flt.get("isHiddenInViewMode", False)
                        filter_type = flt.get("type", "Basic")
                        filter_used = (
                            True if "Where" in flt.get("filter", {}) else False
                        )

                        entity_property_pairs = helper.find_entity_property_pairs(flt)

                        for object_name, properties in entity_property_pairs.items():
                            new_data = {
                                "Page Name": page_id,
                                "Page Display Name": page_display,
                                "Filter Name": filter_name,
                                "Type": filter_type,
                                "Table Name": properties[0],
                                "Object Name": object_name,
                                "Object Type": properties[1],
                                "Hidden": hidden,
                                "Locked": locked,
                                "How Created": how_created,
                                "Used": filter_used,
                            }

                            df = pd.concat(
                                [df, pd.DataFrame(new_data, index=[0])],
                                ignore_index=True,
                            )

        df["Page URL"] = df["Page Name"].apply(
            lambda page_name: f"{helper.get_web_url(report=self._report, workspace=self._workspace)}/{page_name}"
        )

        bool_cols = ["Hidden", "Locked", "Used"]
        df[bool_cols] = df[bool_cols].astype(bool)

        if extended:
            df = self._add_extended(dataframe=df)

        return df
        # return df.style.format({"Page URL": _make_clickable})

    def list_visual_filters(self, extended: bool = False) -> pd.DataFrame:
        """
        Shows a list of all visual filters used in the report.

        Parameters
        ----------
        extended : bool, default=False
            If True, adds an extra column called 'Valid Semantic Model Object' which identifies whether the semantic model object used
            in the report exists in the semantic model which feeds data to the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all the visual filters used in the report.
        """

        rd = self.rdef
        df = pd.DataFrame(
            columns=[
                "Page Name",
                "Page Display Name",
                "Visual Name",
                "Filter Name",
                "Type",
                "Table Name",
                "Object Name",
                "Object Type",
                "Hidden",
                "Locked",
                "How Created",
                "Used",
            ]
        )
        page_mapping, visual_mapping = helper.visual_page_mapping(self)

        for _, r in rd.iterrows():
            path = r["path"]
            payload = r["payload"]
            if path.endswith("/visual.json"):
                obj_file = base64.b64decode(payload).decode("utf-8")
                obj_json = json.loads(obj_file)
                page_id = visual_mapping.get(path)[0]
                page_display = visual_mapping.get(path)[1]
                visual_name = obj_json.get("name")

                if "filterConfig" in obj_json:
                    for flt in obj_json.get("filterConfig", {}).get("filters", {}):
                        filter_name = flt.get("name")
                        how_created = flt.get("howCreated")
                        locked = flt.get("isLockedInViewMode", False)
                        hidden = flt.get("isHiddenInViewMode", False)
                        filter_type = flt.get("type", "Basic")
                        filter_used = (
                            True if "Where" in flt.get("filter", {}) else False
                        )

                        entity_property_pairs = helper.find_entity_property_pairs(flt)

                        for object_name, properties in entity_property_pairs.items():
                            new_data = {
                                "Page Name": page_id,
                                "Page Display Name": page_display,
                                "Visual Name": visual_name,
                                "Filter Name": filter_name,
                                "Type": filter_type,
                                "Table Name": properties[0],
                                "Object Name": object_name,
                                "Object Type": properties[1],
                                "Hidden": hidden,
                                "Locked": locked,
                                "How Created": how_created,
                                "Used": filter_used,
                            }

                            df = pd.concat(
                                [df, pd.DataFrame(new_data, index=[0])],
                                ignore_index=True,
                            )

        bool_cols = ["Hidden", "Locked", "Used"]
        df[bool_cols] = df[bool_cols].astype(bool)

        if extended:
            df = self._add_extended(dataframe=df)

        return df

    def list_visual_interactions(self) -> pd.DataFrame:
        """
        Shows a list of all modified `visual interactions <https://learn.microsoft.com/power-bi/create-reports/service-reports-visual-interactions?tabs=powerbi-desktop>`_ used in the report.

        Parameters
        ----------

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all modified visual interactions used in the report.
        """

        rd = self.rdef
        df = pd.DataFrame(
            columns=[
                "Page Name",
                "Page Display Name",
                "Source Visual Name",
                "Target Visual Name",
                "Type",
            ]
        )

        for _, r in rd.iterrows():
            file_path = r["path"]
            payload = r["payload"]
            if file_path.endswith("/page.json"):
                obj_file = base64.b64decode(payload).decode("utf-8")
                obj_json = json.loads(obj_file)
                page_name = obj_json.get("name")
                page_display = obj_json.get("displayName")

                for vizInt in obj_json.get("visualInteractions", []):
                    sourceVisual = vizInt.get("source")
                    targetVisual = vizInt.get("target")
                    vizIntType = vizInt.get("type")

                    new_data = {
                        "Page Name": page_name,
                        "Page Display Name": page_display,
                        "Source Visual Name": sourceVisual,
                        "Target Visual Name": targetVisual,
                        "Type": vizIntType,
                    }
                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )

        return df

    def list_pages(self) -> pd.DataFrame:
        """
        Shows a list of all pages in the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all pages in the report.
        """

        rd = self.rdef
        df = pd.DataFrame(
            columns=[
                "File Path",
                "Page Name",
                "Page Display Name",
                "Hidden",
                "Active",
                "Width",
                "Height",
                "Display Option",
                "Type",
                "Alignment",
                "Drillthrough Target Page",
                "Visual Count",
                "Data Visual Count",
                "Visible Visual Count",
                "Page Filter Count",
            ]
        )

        dfV = self.list_visuals()

        page_rows = rd[rd["path"].str.endswith("/page.json")]
        pages_row = rd[rd["path"] == "definition/pages/pages.json"]

        for _, r in page_rows.iterrows():
            file_path = r["path"]
            payload = r["payload"]

            pageFile = base64.b64decode(payload).decode("utf-8")
            page_prefix = file_path[0:-9]
            pageJson = json.loads(pageFile)
            page_name = pageJson.get("name")
            height = pageJson.get("height")
            width = pageJson.get("width")

            # Alignment
            matches = parse(
                "$.objects.displayArea[0].properties.verticalAlignment.expr.Literal.Value"
            ).find(pageJson)
            alignment_value = (
                matches[0].value[1:-1] if matches and matches[0].value else "Top"
            )

            # Drillthrough
            matches = parse("$.filterConfig.filters[*].howCreated").find(pageJson)
            how_created_values = [match.value for match in matches]
            drill_through = any(value == "Drillthrough" for value in how_created_values)
            # matches = parse("$.filterConfig.filters[*]").find(pageJson)
            # drill_through = any(
            #    filt.get("howCreated") == "Drillthrough"
            #    for filt in (match.value for match in matches)
            # )

            visual_count = len(
                rd[
                    rd["path"].str.endswith("/visual.json")
                    & (rd["path"].str.startswith(page_prefix))
                ]
            )
            data_visual_count = len(
                dfV[(dfV["Page Name"] == page_name) & (dfV["Data Visual"])]
            )
            visible_visual_count = len(
                dfV[(dfV["Page Name"] == page_name) & (dfV["Hidden"] == False)]
            )

            # Page Filter Count
            matches = parse("$.filterConfig.filters").find(pageJson)
            page_filter_count = (
                len(matches[0].value) if matches and matches[0].value else 0
            )

            # Hidden
            matches = parse("$.visibility").find(pageJson)
            is_hidden = any(match.value == "HiddenInViewMode" for match in matches)

            new_data = {
                "File Path": file_path,
                "Page Name": page_name,
                "Page Display Name": pageJson.get("displayName"),
                "Display Option": pageJson.get("displayOption"),
                "Height": height,
                "Width": width,
                "Hidden": is_hidden,
                "Active": False,
                "Type": helper.page_type_mapping.get((width, height), "Custom"),
                "Alignment": alignment_value,
                "Drillthrough Target Page": drill_through,
                "Visual Count": visual_count,
                "Data Visual Count": data_visual_count,
                "Visible Visual Count": visible_visual_count,
                "Page Filter Count": page_filter_count,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

        page_payload = pages_row["payload"].iloc[0]
        pageFile = base64.b64decode(page_payload).decode("utf-8")
        pageJson = json.loads(pageFile)
        activePage = pageJson["activePageName"]

        df.loc[df["Page Name"] == activePage, "Active"] = True

        int_cols = [
            "Width",
            "Height",
            "Page Filter Count",
            "Visual Count",
            "Visible Visual Count",
            "Data Visual Count",
        ]
        df[int_cols] = df[int_cols].astype(int)

        bool_cols = ["Hidden", "Active", "Drillthrough Target Page"]
        df[bool_cols] = df[bool_cols].astype(bool)

        df["Page URL"] = df["Page Name"].apply(
            lambda page_name: f"{helper.get_web_url(report=self._report, workspace=self._workspace)}/{page_name}"
        )

        return df
        # return df.style.format({"Page URL": _make_clickable})

    def list_visuals(self) -> pd.DataFrame:
        """
        Shows a list of all visuals in the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all visuals in the report.
        """

        rd = self.rdef
        df = pd.DataFrame(
            columns=[
                "File Path",
                "Page Name",
                "Page Display Name",
                "Visual Name",
                "Type",
                "Display Type",
                "X",
                "Y",
                "Z",
                "Width",
                "Height",
                "Tab Order",
                "Hidden",
                "Title",
                "SubTitle",
                "Custom Visual",
                "Alt Text",
                "Show Items With No Data",
                "Divider",
                "Slicer Type",
                "Row SubTotals",
                "Column SubTotals",
                "Data Visual",
                "Has Sparkline",
                "Visual Filter Count",
                "Data Limit",
            ]
        )

        rd_filt = rd[rd["path"] == "definition/report.json"]
        payload = rd_filt["payload"].iloc[0]
        rptJson = _extract_json(rd_filt)
        custom_visuals = rptJson.get("publicCustomVisuals", [])
        page_mapping, visual_mapping = helper.visual_page_mapping(self)
        helper.populate_custom_visual_display_names()

        def contains_key(data, keys_to_check):
            matches = parse("$..*").find(data)

            all_keys = set()
            for match in matches:
                if isinstance(match.value, dict):
                    all_keys.update(match.value.keys())
                elif isinstance(match.value, list):
                    for item in match.value:
                        if isinstance(item, dict):
                            all_keys.update(item.keys())

            return any(key in all_keys for key in keys_to_check)

        for _, r in rd.iterrows():
            file_path = r["path"]
            payload = r["payload"]
            if file_path.endswith("/visual.json"):
                visual_file = base64.b64decode(payload).decode("utf-8")
                visual_json = json.loads(visual_file)
                page_id = visual_mapping.get(file_path)[0]
                page_display = visual_mapping.get(file_path)[1]
                pos = visual_json.get("position")

                # Visual Type
                matches = parse("$.visual.visualType").find(visual_json)
                visual_type = matches[0].value if matches else "Group"

                visual_type_display = helper.vis_type_mapping.get(
                    visual_type, visual_type
                )
                cst_value, rst_value, slicer_type = False, False, "N/A"

                # Visual Filter Count
                matches = parse("$.filterConfig.filters[*]").find(visual_json)
                visual_filter_count = len(matches)

                # Data Limit
                matches = parse(
                    '$.filterConfig.filters[?(@.type == "VisualTopN")].filter.Where[*].Condition.VisualTopN.ItemCount'
                ).find(visual_json)
                data_limit = matches[0].value if matches else 0

                # Title
                matches = parse(
                    "$.visual.visualContainerObjects.title[0].properties.text.expr.Literal.Value"
                ).find(visual_json)
                title = matches[0].value[1:-1] if matches else ""

                # SubTitle
                matches = parse(
                    "$.visual.visualContainerObjects.subTitle[0].properties.text.expr.Literal.Value"
                ).find(visual_json)
                sub_title = matches[0].value[1:-1] if matches else ""

                # Alt Text
                matches = parse(
                    "$.visual.visualContainerObjects.general[0].properties.altText.expr.Literal.Value"
                ).find(visual_json)
                alt_text = matches[0].value[1:-1] if matches else ""

                # Show items with no data
                def find_show_all_with_jsonpath(obj):
                    matches = parse("$..showAll").find(obj)
                    return any(match.value is True for match in matches)

                show_all_data = find_show_all_with_jsonpath(visual_json)

                # Divider
                matches = parse(
                    "$.visual.visualContainerObjects.divider[0].properties.show.expr.Literal.Value"
                ).find(visual_json)
                divider = matches[0] if matches else ""

                # Row/Column Subtotals
                if visual_type == "pivotTable":
                    cst_matches = parse(
                        "$.visual.objects.subTotals[0].properties.columnSubtotals.expr.Literal.Value"
                    ).find(visual_json)
                    rst_matches = parse(
                        "$.visual.objects.subTotals[0].properties.rowSubtotals.expr.Literal.Value"
                    ).find(visual_json)

                    if cst_matches:
                        cst_value = False if cst_matches[0].value == "false" else True

                    if rst_matches:
                        rst_value = False if rst_matches[0].value == "false" else True

                # Slicer Type
                if visual_type == "slicer":
                    matches = parse(
                        "$.visual.objects.data[0].properties.mode.expr.Literal.Value"
                    ).find(visual_json)
                    slicer_type = matches[0].value[1:-1] if matches else "N/A"

                # Data Visual
                is_data_visual = contains_key(
                    visual_json,
                    [
                        "Aggregation",
                        "Column",
                        "Measure",
                        "HierarchyLevel",
                        "NativeVisualCalculation",
                    ],
                )

                # Sparkline
                has_sparkline = contains_key(visual_json, ["SparklineData"])

                new_data = {
                    "File Path": file_path,
                    "Page Name": page_id,
                    "Page Display Name": page_display,
                    "Visual Name": visual_json.get("name"),
                    "X": pos.get("x"),
                    "Y": pos.get("y"),
                    "Z": pos.get("z"),
                    "Width": pos.get("width"),
                    "Height": pos.get("height"),
                    "Tab Order": pos.get("tabOrder"),
                    "Hidden": visual_json.get("isHidden", False),
                    "Type": visual_type,
                    "Display Type": visual_type_display,
                    "Title": title,
                    "SubTitle": sub_title,
                    "Custom Visual": visual_type in custom_visuals,
                    "Alt Text": alt_text,
                    "Show Items With No Data": show_all_data,
                    "Divider": divider,
                    "Row SubTotals": rst_value,
                    "Column SubTotals": cst_value,
                    "Slicer Type": slicer_type,
                    "Data Visual": is_data_visual,
                    "Has Sparkline": has_sparkline,
                    "Visual Filter Count": visual_filter_count,
                    "Data Limit": data_limit,
                }

                df = pd.concat(
                    [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                )

        bool_cols = [
            "Hidden",
            "Show Items With No Data",
            "Custom Visual",
            "Data Visual",
            "Has Sparkline",
            "Row SubTotals",
            "Column SubTotals",
        ]

        grouped_df = (
            self.list_visual_objects()
            .groupby(["Page Name", "Visual Name"])
            .size()
            .reset_index(name="Visual Object Count")
        )

        df = pd.merge(
            df,
            grouped_df,
            left_on=["Page Name", "Visual Name"],
            right_on=["Page Name", "Visual Name"],
            how="left",
        )
        df["Visual Object Count"] = df["Visual Object Count"].fillna(0).astype(int)

        float_cols = ["X", "Y", "Width", "Height"]
        int_cols = ["Z", "Visual Filter Count", "Data Limit", "Visual Object Count"]
        df[bool_cols] = df[bool_cols].astype(bool)
        df[int_cols] = df[int_cols].astype(int)
        df[float_cols] = df[float_cols].astype(float)

        return df

    def list_visual_objects(self, extended: bool = False) -> pd.DataFrame:
        """
        Shows a list of all semantic model objects used in each visual in the report.

        Parameters
        ----------
        extended : bool, default=False
            If True, adds an extra column called 'Valid Semantic Model Object' which identifies whether the semantic model object used
            in the report exists in the semantic model which feeds data to the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all semantic model objects used in each visual in the report.
        """

        rd = self.rdef
        page_mapping, visual_mapping = helper.visual_page_mapping(self)
        df = pd.DataFrame(
            columns=[
                "Page Name",
                "Page Display Name",
                "Visual Name",
                "Table Name",
                "Object Name",
                "Object Type",
                "Implicit Measure",
                "Sparkline",
                "Visual Calc",
                "Format",
            ]
        )

        def contains_key(data, keys_to_check):
            if isinstance(data, dict):
                for key, value in data.items():
                    if key in keys_to_check:
                        return True
                    if contains_key(value, keys_to_check):
                        return True
            elif isinstance(data, list):
                for item in data:
                    if contains_key(item, keys_to_check):
                        return True
            return False

        def find_entity_property_pairs(data, result=None, keys_path=None):
            if result is None:
                result = {}
            if keys_path is None:
                keys_path = []

            if isinstance(data, dict):
                if (
                    "Entity" in data.get("Expression", {}).get("SourceRef", {})
                    and "Property" in data
                ):
                    entity = (
                        data.get("Expression", {})
                        .get("SourceRef", {})
                        .get("Entity", {})
                    )
                    property_value = data.get("Property", {})
                    object_type = keys_path[-1].replace("HierarchyLevel", "Hierarchy")
                    is_agg = keys_path[-3] == "Aggregation"
                    is_viz_calc = keys_path[-3] == "NativeVisualCalculation"
                    is_sparkline = keys_path[-3] == "SparklineData"
                    result[property_value] = (
                        entity,
                        object_type,
                        is_agg,
                        is_viz_calc,
                        is_sparkline,
                    )
                    keys_path.pop()

                # Recursively search the rest of the dictionary
                for key, value in data.items():
                    keys_path.append(key)
                    find_entity_property_pairs(value, result, keys_path)

            elif isinstance(data, list):
                for item in data:
                    find_entity_property_pairs(item, result, keys_path)

            return result

        for _, r in rd.iterrows():
            file_path = r["path"]
            payload = r["payload"]
            if file_path.endswith("/visual.json"):
                visual_file = base64.b64decode(payload).decode("utf-8")
                visual_json = json.loads(visual_file)
                page_id = visual_mapping.get(file_path)[0]
                page_display = visual_mapping.get(file_path)[1]

                entity_property_pairs = find_entity_property_pairs(visual_json)
                query_state = (
                    visual_json.get("visual", {})
                    .get("query", {})
                    .get("queryState", {})
                    .get("Values", {})
                )
                format_mapping = {}
                for p in query_state.get("projections", []):
                    query_ref = p.get("queryRef")
                    fmt = p.get("format")
                    if fmt is not None:
                        format_mapping[query_ref] = fmt

                for object_name, properties in entity_property_pairs.items():
                    table_name = properties[0]
                    obj_full = f"{table_name}.{object_name}"
                    is_agg = properties[2]
                    format_value = format_mapping.get(obj_full)

                    if is_agg:
                        for k, v in format_mapping.items():
                            if obj_full in k:
                                format_value = v
                    new_data = {
                        "Page Name": page_id,
                        "Page Display Name": page_display,
                        "Visual Name": visual_json.get("name"),
                        "Table Name": table_name,
                        "Object Name": object_name,
                        "Object Type": properties[1],
                        "Implicit Measure": is_agg,
                        "Sparkline": properties[4],
                        "Visual Calc": properties[3],
                        "Format": format_value,
                    }

                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )

        if extended:
            df = self._add_extended(dataframe=df)

        bool_cols = ["Implicit Measure", "Sparkline", "Visual Calc"]
        df[bool_cols] = df[bool_cols].astype(bool)

        return df

    def list_semantic_model_objects(self, extended: bool = False) -> pd.DataFrame:
        """
        Shows a list of all semantic model objects (measures, columns, hierarchies) that are used in the report and where the objects
        were used (i.e. visual, report filter, page filter, visual filter).

        Parameters
        ----------
        extended : bool, default=False
            If True, adds an extra column called 'Valid Semantic Model Object' which identifies whether the semantic model object used
            in the report exists in the semantic model which feeds data to the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe showing the semantic model objects used in the report.
        """

        from sempy_labs.tom import connect_semantic_model

        df = pd.DataFrame(
            columns=[
                "Table Name",
                "Object Name",
                "Object Type",
                "Report Source",
                "Report Source Object",
            ]
        )

        rf = self.list_report_filters()
        pf = self.list_page_filters()
        vf = self.list_visual_filters()
        vo = self.list_visual_objects()

        rf_subset = rf[["Table Name", "Object Name", "Object Type"]].copy()
        rf_subset["Report Source"] = "Report Filter"
        rf_subset["Report Source Object"] = self._report

        pf_subset = pf[
            ["Table Name", "Object Name", "Object Type", "Page Display Name"]
        ].copy()
        pf_subset["Report Source"] = "Page Filter"
        pf_subset["Report Source Object"] = pf_subset["Page Display Name"]
        pf_subset.drop(columns=["Page Display Name"], inplace=True)

        vf_subset = vf[
            [
                "Table Name",
                "Object Name",
                "Object Type",
                "Page Display Name",
                "Visual Name",
            ]
        ].copy()
        vf_subset["Report Source"] = "Visual Filter"
        vf_subset["Report Source Object"] = format_dax_object_name(
            vf_subset["Page Display Name"], vf_subset["Visual Name"]
        )
        vf_subset.drop(columns=["Page Display Name", "Visual Name"], inplace=True)

        vo_subset = vo[
            [
                "Table Name",
                "Object Name",
                "Object Type",
                "Page Display Name",
                "Visual Name",
            ]
        ].copy()
        vo_subset["Report Source"] = "Visual"
        vo_subset["Report Source Object"] = format_dax_object_name(
            vo_subset["Page Display Name"], vo_subset["Visual Name"]
        )
        vo_subset.drop(columns=["Page Display Name", "Visual Name"], inplace=True)

        df = pd.concat(
            [df, rf_subset, pf_subset, vf_subset, vo_subset], ignore_index=True
        )

        if extended:
            dataset_id, dataset_name, dataset_workspace_id, dataset_workspace = (
                resolve_dataset_from_report(
                    report=self._report, workspace=self._workspace
                )
            )

            def check_validity(tom, row):
                object_validators = {
                    "Measure": lambda: any(
                        o.Name == row["Object Name"] for o in tom.all_measures()
                    ),
                    "Column": lambda: any(
                        format_dax_object_name(c.Parent.Name, c.Name)
                        == format_dax_object_name(row["Table Name"], row["Object Name"])
                        for c in tom.all_columns()
                    ),
                    "Hierarchy": lambda: any(
                        format_dax_object_name(h.Parent.Name, h.Name)
                        == format_dax_object_name(row["Table Name"], row["Object Name"])
                        for h in tom.all_hierarchies()
                    ),
                }
                return object_validators.get(row["Object Type"], lambda: False)()

            with connect_semantic_model(
                dataset=dataset_name, readonly=True, workspace=dataset_workspace
            ) as tom:
                df["Valid Semantic Model Object"] = df.apply(
                    lambda row: check_validity(tom, row), axis=1
                )

        return df

    def _list_all_semantic_model_objects(self):

        # Includes dependencies

        df = (
            self.list_semantic_model_objects()[
                ["Table Name", "Object Name", "Object Type"]
            ]
            .drop_duplicates()
            .reset_index(drop=True)
        )
        dataset_id, dataset_name, dataset_workspace_id, dataset_workspace = (
            resolve_dataset_from_report(report=self._report, workspace=self._workspace)
        )
        dep = get_measure_dependencies(
            dataset=dataset_name, workspace=dataset_workspace
        )
        rpt_measures = df[df["Object Type"] == "Measure"]["Object Name"].values
        new_rows = dep[dep["Object Name"].isin(rpt_measures)][
            ["Referenced Table", "Referenced Object", "Referenced Object Type"]
        ]
        new_rows.columns = ["Table Name", "Object Name", "Object Type"]
        result_df = (
            pd.concat([df, new_rows], ignore_index=True)
            .drop_duplicates()
            .reset_index(drop=True)
        )

        result_df["Dataset Name"] = dataset_name
        result_df["Dataset Workspace Name"] = dataset_workspace
        colName = "Dataset Name"
        result_df.insert(0, colName, result_df.pop(colName))
        colName = "Dataset Workspace Name"
        result_df.insert(1, colName, result_df.pop(colName))

        return result_df

    def list_bookmarks(self) -> pd.DataFrame:
        """
        Shows a list of all bookmarks in the report.

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all bookmarks in the report.
        """

        rd = self.rdef
        df = pd.DataFrame(
            columns=[
                "File Path",
                "Bookmark Name",
                "Bookmark Display Name",
                "Page Name",
                "Page Display Name",
                "Visual Name",
                "Visual Hidden",
            ]
        )

        bookmark_rows = rd[rd["path"].str.endswith(".bookmark.json")]

        for _, r in bookmark_rows.iterrows():
            path = r["path"]
            payload = r["payload"]

            obj_file = base64.b64decode(payload).decode("utf-8")
            obj_json = json.loads(obj_file)

            bookmark_name = obj_json.get("name")
            bookmark_display = obj_json.get("displayName")
            rpt_page_id = obj_json.get("explorationState", {}).get("activeSection")
            page_id, page_display, file_path = helper.resolve_page_name(
                self, page_name=rpt_page_id
            )

            for rptPg in obj_json.get("explorationState", {}).get("sections", {}):
                for visual_name in (
                    obj_json.get("explorationState", {})
                    .get("sections", {})
                    .get(rptPg, {})
                    .get("visualContainers", {})
                ):
                    if (
                        obj_json.get("explorationState", {})
                        .get("sections", {})
                        .get(rptPg, {})
                        .get("visualContainers", {})
                        .get(visual_name, {})
                        .get("singleVisual", {})
                        .get("display", {})
                        .get("mode", {})
                        == "hidden"
                    ):
                        visual_hidden = True
                    else:
                        visual_hidden = False

                    new_data = {
                        "File Path": path,
                        "Bookmark Name": bookmark_name,
                        "Bookmark Display Name": bookmark_display,
                        "Page Name": page_id,
                        "Page Display Name": page_display,
                        "Visual Name": visual_name,
                        "Visual Hidden": visual_hidden,
                    }
                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )

        bool_cols = ["Visual Hidden"]
        df[bool_cols] = df[bool_cols].astype(bool)

        return df

    def list_report_level_measures(self) -> pd.DataFrame:
        """
        Shows a list of all `report-level measures <https://learn.microsoft.com/power-bi/transform-model/desktop-measures#report-level-measures>`_ in the report.

        Parameters
        ----------

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe containing a list of all report-level measures in the report.
        """

        df = pd.DataFrame(
            columns=[
                "Measure Name",
                "Table Name",
                "Expression",
                "Data Type",
                "Format String",
            ]
        )
        rd = self.rdef
        rd_filt = rd[rd["path"] == "definition/reportExtensions.json"]

        if len(rd_filt) == 1:
            payload = rd_filt["payload"].iloc[0]
            obj_file = base64.b64decode(payload).decode("utf-8")
            obj_json = json.loads(obj_file)

            for e in obj_json.get("entities", []):
                table_name = e.get("name")
                for m in e.get("measures", []):
                    measure_name = m.get("name")
                    expr = m.get("expression")
                    data_type = m.get("dataType")
                    format_string = m.get("formatString")

                    new_data = {
                        "Measure Name": measure_name,
                        "Table Name": table_name,
                        "Expression": expr,
                        "Data Type": data_type,
                        "Format String": format_string,
                    }
                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )

        return df

    # Automation functions
    def set_theme(self, theme_file_path: str):
        """
        Sets a custom theme for a report based on a theme .json file.

        Parameters
        ----------
        theme_file_path : str
            The file path of the theme.json file. This can either be from a Fabric lakehouse or from the web.
            Example for lakehouse: file_path = '/lakehouse/default/Files/CY23SU09.json'
            Example for web url: file_path = 'https://raw.githubusercontent.com/PowerBiDevCamp/FabricUserApiDemo/main/FabricUserApiDemo/DefinitionTemplates/Shared/Reports/StaticResources/SharedResources/BaseThemes/CY23SU08.json'
        """

        import requests

        report_path = "definition/report.json"
        theme_version = "5.5.4"
        request_body = {"definition": {"parts": []}}

        if not theme_file_path.endswith(".json"):
            raise ValueError(
                f"{icons.red_dot} The '{theme_file_path}' theme file path must be a .json file."
            )
        elif theme_file_path.startswith("https://"):
            response = requests.get(theme_file_path)
            json_file = response.json()
        elif theme_file_path.startswith("/lakehouse"):
            with open(theme_file_path, "r", encoding="utf-8-sig") as file:
                json_file = json.load(file)
        else:
            ValueError(
                f"{icons.red_dot} Incorrect theme file path value '{theme_file_path}'."
            )

        theme_name = json_file["name"]
        theme_name_full = f"{theme_name}.json"

        # Add theme.json file to request_body
        file_payload = _conv_b64(json_file)
        filePath = f"StaticResources/RegisteredResources/{theme_name_full}"

        _add_part(request_body, filePath, file_payload)

        new_theme = {
            "name": theme_name_full,
            "path": theme_name_full,
            "type": "CustomTheme",
        }

        rd = self.rdef
        for _, r in rd.iterrows():
            path = r["path"]
            payload = r["payload"]
            if path != report_path:
                _add_part(request_body, path, payload)
            # Update the report.json file
            else:
                rptFile = base64.b64decode(payload).decode("utf-8")
                rptJson = json.loads(rptFile)
                resource_type = "RegisteredResources"

                # Add to theme collection
                if "customTheme" not in rptJson["themeCollection"]:
                    rptJson["themeCollection"]["customTheme"] = {
                        "name": theme_name_full,
                        "reportVersionAtImport": theme_version,
                        "type": resource_type,
                    }
                else:
                    rptJson["themeCollection"]["customTheme"]["name"] = theme_name_full
                    rptJson["themeCollection"]["customTheme"]["type"] = resource_type

                for package in rptJson["resourcePackages"]:
                    package["items"] = [
                        item
                        for item in package["items"]
                        if item["type"] != "CustomTheme"
                    ]

                if not any(
                    package["name"] == resource_type
                    for package in rptJson["resourcePackages"]
                ):
                    new_registered_resources = {
                        "name": resource_type,
                        "type": resource_type,
                        "items": [new_theme],
                    }
                    rptJson["resourcePackages"].append(new_registered_resources)
                else:
                    names = [
                        rp["name"] for rp in rptJson["resourcePackages"][1]["items"]
                    ]

                    if theme_name_full not in names:
                        rptJson["resourcePackages"][1]["items"].append(new_theme)

                file_payload = _conv_b64(rptJson)
                _add_part(request_body, path, file_payload)

        self.update_report(request_body=request_body)
        print(
            f"{icons.green_dot} The '{theme_name}' theme has been set as the theme for the '{self._report}' report within the '{self._workspace}' workspace."
        )

    def set_active_page(self, page_name: str):
        """
        Sets the active page (first page displayed when opening a report) for a report.

        Parameters
        ----------
        page_name : str
            The page name or page display name of the report.
        """

        pages_file = "definition/pages/pages.json"
        request_body = {"definition": {"parts": []}}

        rd = self.rdef
        page_id, page_display, file_path = helper.resolve_page_name(
            self, page_name=page_name
        )
        for _, r in rd.iterrows():
            path = r["path"]
            file_payload = r["payload"]
            if path != pages_file:
                _add_part(request_body, path, file_payload)

        pagePath = rd[rd["path"] == pages_file]
        payload = pagePath["payload"].iloc[0]
        pageFile = base64.b64decode(payload).decode("utf-8")
        pageJson = json.loads(pageFile)
        pageJson["activePageName"] = page_id
        file_payload = _conv_b64(pageJson)

        _add_part(request_body, pages_file, file_payload)

        self.update_report(request_body=request_body)
        print(
            f"{icons.green_dot} The '{page_name}' page has been set as the active page in the '{self._report}' report within the '{self._workspace}' workspace."
        )

    def set_page_type(self, page_name: str, page_type: str):

        if page_type not in helper.page_types:
            raise ValueError(
                f"{icons.red_dot} Invalid page type. Valid options: {helper.page_types}."
            )

        request_body = {"definition": {"parts": []}}

        letter_key = next(
            (
                key
                for key, value in helper.page_type_mapping.items()
                if value == page_type
            ),
            None,
        )
        if letter_key:
            width, height = letter_key
        else:
            raise ValueError(
                "Invalid page_type parameter. Valid options: ['Tooltip', 'Letter', '4:3', '16:9']."
            )

        rd = self.rdef
        page_id, page_display, file_path = helper.resolve_page_name(
            self, page_name=page_name
        )
        rd_filt = rd[rd["path"] == file_path]
        payload = rd_filt["payload"].iloc[0]
        pageFile = base64.b64decode(payload).decode("utf-8")
        pageJson = json.loads(pageFile)
        pageJson["width"] = width
        pageJson["height"] = height

        file_payload = _conv_b64(pageJson)
        _add_part(request_body, file_path, file_payload)

        for _, r in rd.iterrows():
            if r["path"] != file_path:
                _add_part(request_body, r["path"], r["payload"])

        self.update_report(request_body=request_body)
        print(
            f"The '{page_display}' page has been updated to the '{page_type}' page type."
        )

    def remove_unnecessary_custom_visuals(self):
        """
        Removes any custom visuals within the report that are not used in the report.
        """

        dfCV = self.list_custom_visuals()
        dfV = self.list_visuals()
        rd = self.rdef
        cv_remove = []
        cv_remove_display = []
        request_body = {"definition": {"parts": []}}

        for _, r in dfCV.iterrows():
            cv = r["Custom Visual Name"]
            cv_display = r["Custom Visual Display Name"]
            dfV_filt = dfV[dfV["Type"] == cv]
            if len(dfV_filt) == 0:
                cv_remove.append(cv)  # Add to the list for removal
                cv_remove_display.append(cv_display)
        if len(cv_remove) == 0:
            print(
                f"{icons.green_dot} There are no unnecessary custom visuals in the '{self._report}' report within the '{self._workspace}' workspace."
            )
            return

        for _, r in rd.iterrows():
            file_path = r["path"]
            payload = r["payload"]
            if file_path == "definition/report.json":
                rpt_file = base64.b64decode(payload).decode("utf-8")
                rpt_json = json.loads(rpt_file)
                rpt_json["publicCustomVisuals"] = [
                    item
                    for item in rpt_json["publicCustomVisuals"]
                    if item not in cv_remove
                ]

                payload = _conv_b64(rpt_json)

            _add_part(request_body, file_path, payload)

        self.update_report(request_body=request_body)
        print(
            f"{icons.green_dot} The {cv_remove_display} custom visuals have been removed from the '{self._report}' report within the '{self._workspace}' workspace."
        )

    def migrate_report_level_measures(self, measures: Optional[str | List[str]] = None):
        """
        Moves all report-level measures from the report to the semantic model on which the report is based.

        Parameters
        ----------
        measures : str | List[str], default=None
            A measure or list of measures to move to the semantic model.
            Defaults to None which resolves to moving all report-level measures to the semantic model.
        """

        from sempy_labs.tom import connect_semantic_model

        rlm = self.list_report_level_measures()
        if len(rlm) == 0:
            print(
                f"{icons.green_dot} The '{self._report}' report within the '{self._workspace}' workspace has no report-level measures."
            )
            return

        dataset_id, dataset_name, dataset_workspace_id, dataset_workspace = (
            resolve_dataset_from_report(report=self._report, workspace=self._workspace)
        )

        if isinstance(measures, str):
            measures = [measures]

        request_body = {"definition": {"parts": []}}
        rpt_file = "definition/reportExtensions.json"

        rd = self.rdef
        rd_filt = rd[rd["path"] == rpt_file]
        payload = rd_filt["payload"].iloc[0]
        extFile = base64.b64decode(payload).decode("utf-8")
        extJson = json.loads(extFile)

        mCount = 0
        with connect_semantic_model(
            dataset=dataset_name, readonly=False, workspace=dataset_workspace
        ) as tom:
            for _, r in rlm.iterrows():
                tableName = r["Table Name"]
                mName = r["Measure Name"]
                mExpr = r["Expression"]
                # mDataType = r["Data Type"]
                mformatString = r["Format String"]
                # Add measures to the model
                if mName in measures or measures is None:
                    tom.add_measure(
                        table_name=tableName,
                        measure_name=mName,
                        expression=mExpr,
                        format_string=mformatString,
                    )
                    tom.set_annotation(
                        object=tom.model.Tables[tableName].Measures[mName],
                        name="semanticlinklabs",
                        value="reportlevelmeasure",
                    )
                mCount += 1
            # Remove measures from the json
            if measures is not None and len(measures) < mCount:
                for e in extJson["entities"]:
                    e["measures"] = [
                        measure
                        for measure in e["measures"]
                        if measure["name"] not in measures
                    ]
                extJson["entities"] = [
                    entity for entity in extJson["entities"] if entity["measures"]
                ]
                file_payload = _conv_b64(extJson)
                _add_part(request_body, rpt_file, file_payload)

        # Add unchanged payloads
        for _, r in rd.iterrows():
            path = r["path"]
            payload = r["payload"]
            if path != rpt_file:
                _add_part(request_body, path, payload)

        self.update_report(request_body=request_body)
        print(
            f"{icons.green_dot} The report-level measures have been migrated to the '{dataset_name}' semantic model within the '{dataset_workspace}' workspace."
        )

    def set_page_visibility(self, page_name: str, hidden: bool):
        """
        Sets whether a report page is visible or hidden.

        Parameters
        ----------
        page_name : str
            The page name or page display name of the report.
        hidden : bool
            If set to True, hides the report page.
            If set to False, makes the report page visible.
        """

        rd = self.rdef
        page_id, page_display, file_path = helper.resolve_page_name(
            self, page_name=page_name
        )
        visibility = "visible" if hidden is False else "hidden"

        request_body = {"definition": {"parts": []}}

        for _, r in rd.iterrows():
            path = r["path"]
            payload = r["payload"]
            if path == file_path:
                obj_file = base64.b64decode(payload).decode("utf-8")
                obj_json = json.loads(obj_file)
                if hidden:
                    obj_json["visibility"] = "HiddenInViewMode"
                else:
                    if "visibility" in obj_json:
                        del obj_json["visibility"]
                _add_part(request_body, path, _conv_b64(obj_json))
            else:
                _add_part(request_body, path, payload)

        self.update_report(request_body=request_body)
        print(f"{icons.green_dot} The '{page_name}' page has been set to {visibility}.")

    def hide_tooltip_drillthrough_pages(self):
        """
        Hides all tooltip pages and drillthrough pages in a report.
        """

        dfP = self.list_pages()
        dfP_filt = dfP[
            (dfP["Type"] == "Tooltip") | (dfP["Drillthrough Target Page"] == True)
        ]

        if len(dfP_filt) == 0:
            print(
                f"{icons.green_dot} There are no Tooltip or Drillthrough pages in the '{self._report}' report within the '{self._workspace}' workspace."
            )
            return

        for _, r in dfP_filt.iterrows():
            page_name = r["Page Name"]
            self.set_page_visibility(page_name=page_name, hidden=True)

    def disable_show_items_with_no_data(self):
        """
        Disables the `show items with no data <https://learn.microsoft.com/power-bi/create-reports/desktop-show-items-no-data>`_ property in all visuals within the report.
        """

        request_body = {"definition": {"parts": []}}

        def delete_key_in_json(obj, key_to_delete):
            if isinstance(obj, dict):
                if key_to_delete in obj:
                    del obj[key_to_delete]
                for key, value in obj.items():
                    delete_key_in_json(value, key_to_delete)
            elif isinstance(obj, list):
                for item in obj:
                    delete_key_in_json(item, key_to_delete)

        rd = self.rdef
        for _, r in rd.iterrows():
            file_path = r["path"]
            payload = r["payload"]
            if file_path.endswith("/visual.json"):
                objFile = base64.b64decode(payload).decode("utf-8")
                objJson = json.loads(objFile)
                delete_key_in_json(objJson, "showAll")
                _add_part(request_body, file_path, _conv_b64(objJson))
            else:
                _add_part(request_body, file_path, payload)

        self.update_report(request_body=request_body)
        print(
            f"{icons.green_dot} Show items with data has been disabled for all visuals in the '{self._report}' report within the '{self._workspace}' workspace."
        )

    def __get_annotation_value(self, object_name: str, object_type: str, name: str):

        object_types = ["Visual", "Page", "Report"]
        object_type = object_type.capitalize()
        if object_type not in object_types:
            raise ValueError(
                f"{icons.red_dot} Invalid object type. Valid options: {object_types}."
            )

        rd = self.rdef

        if object_type == "Report":
            rd_filt = rd[rd["path"] == "definition/report.json"]
            obj_json = _extract_json(rd_filt)
        elif object_type == "Page":
            page_id, page_display, page_file = helper.resolve_page_name(
                self, page_name=object_name
            )
            rd_filt = rd[rd["path"] == page_file]
            payload = rd_filt["payload"].iloc[0]
            obj_file = base64.b64decode(payload).decode("utf-8")
            obj_json = json.loads(obj_file)
        elif object_type == "Visual":
            pattern = r"'([^']+)'\[([^]]+)\]"
            match = re.search(pattern, object_name)
            if match:
                p_name = match.group(1)
                v_name = match.group(2)
            else:
                raise ValueError(
                    "Invalid page/visual name within the 'object_name' parameter. Valid format: 'Page 1'[f8dvo24PdJ39fp6]"
                )
            valid_page_name, valid_display_name, visual_name, file_path = (
                helper.resolve_visual_name(self, page_name=p_name, visual_name=v_name)
            )
            rd_filt = rd[rd["path"] == file_path]
            payload = rd_filt["payload"].iloc[0]
            obj_file = base64.b64decode(payload).decode("utf-8")
            obj_json = json.loads(obj_file)

        value = obj_json.get("annotations", {}).get(name, "")

        return value

    def __remove_annotation(self, object_name: str, object_type: str, name: str):

        object_types = ["Visual", "Page", "Report"]
        object_type = object_type.capitalize()
        if object_type not in object_types:
            raise ValueError(
                f"{icons.red_dot} Invalid object type. Valid options: {object_types}."
            )

    def __set_annotation(
        self, object_name: str, object_type: str, name: str, value: str
    ):

        object_types = ["Visual", "Page", "Report"]
        object_type = object_type.capitalize()
        if object_type not in object_types:
            raise ValueError(
                f"{icons.red_dot} Invalid object type. Valid options: {object_types}."
            )

        request_body = {"definition": {"parts": []}}
        new_annotation = {"name": name, "value": value}

        # Creates the annotation if it does not exist. Updates the annotation value if the annotation already exists
        def update_annotation(payload):
            objFile = base64.b64decode(payload).decode("utf-8")
            objJson = json.loads(objFile)
            if "annotations" not in objJson:
                objJson["annotations"] = [new_annotation]
            else:
                names = []
                for ann in objJson["annotations"]:
                    names.append(ann["name"])
                if name not in names:
                    objJson["annotations"].append(new_annotation)
                else:
                    for ann in objJson["annotations"]:
                        if ann["name"] == name:
                            ann["value"] = value
            return objJson

        # Validate page and visual names
        if object_type == "Page":
            page_id, page_display, file_path = helper.resolve_page_name(
                self, page_name=object_name
            )
        elif object_type == "Visual":
            pattern = r"'(.*?)'|\[(.*?)\]"
            matches = re.findall(pattern, object_name)
            page_name = matches[0][0]
            visual_id = matches[1][1]
            page_id, page_display, file_path = helper.resolve_page_name(
                self, page_name=page_name
            )

        rd = self.rdef
        for _, r in rd.iterrows():
            path = r["path"]
            payload = r["payload"]
            if object_type == "Report" and path == "definition/report.json":
                a = update_annotation(payload=payload)
                _add_part(request_body, path, _conv_b64(a))
            elif (
                object_type == "Page"
                and path == f"definition/pages/{page_id}/page.json"
            ):
                a = update_annotation(payload=payload)
                _add_part(request_body, path, _conv_b64(a))
            elif (
                object_type == "Visual"
                and path
                == f"definition/pages/{page_id}/visuals/{visual_id}/visual.json"
            ):
                a = update_annotation(payload=payload)
                _add_part(request_body, path, _conv_b64(a))
            else:
                _add_part(request_body, path, payload)

        self.update_report(request_body=request_body)
        if object_type == "Report":
            print(
                f"{icons.green_dot} The '{name}' annotation has been set on the report with the '{value}' value."
            )
        elif object_type == "Page":
            print(
                f"{icons.green_dot} The '{name}' annotation has been set on the '{object_name}' page with the '{value}' value."
            )
        elif object_type == "Visual":
            print(
                f"{icons.green_dot} The '{name}' annotation has been set on the '{visual_id}' visual on the '{page_display}' page with the '{value}' value."
            )

    def __adjust_settings(
        self, setting_type: str, setting_name: str, setting_value: bool
    ):  # Meta function

        valid_setting_types = ["settings", "slowDataSourceSettings"]
        valid_settings = [
            "isPersistentUserStateDisabled",
            "hideVisualContainerHeader",
            "defaultFilterActionIsDataFilter",
            "useStylableVisualContainerHeader",
            "useDefaultAggregateDisplayName",
            "useEnhancedTooltips",
            "allowChangeFilterTypes",
            "disableFilterPaneSearch",
            "useCrossReportDrillthrough",
        ]
        valid_slow_settings = [
            "isCrossHighlightingDisabled",
            "isSlicerSelectionsButtonEnabled",
        ]

        if setting_type not in valid_setting_types:
            raise ValueError(
                f"Invalid setting_type. Valid options: {valid_setting_types}."
            )
        if setting_type == "settings" and setting_name not in valid_settings:
            raise ValueError(
                f"The '{setting_name}' is not a valid setting. Valid options: {valid_settings}."
            )
        if (
            setting_type == "slowDataSourceSettings"
            and setting_name not in valid_slow_settings
        ):
            raise ValueError(
                f"The '{setting_name}' is not a valid setting. Valid options: {valid_slow_settings}."
            )

        request_body = {"definition": {"parts": []}}

        rd = self.rdef
        for _, r in rd.iterrows():
            path = r["path"]
            payload = r["payload"]
            if path == "definition/report.json":
                obj_file = base64.b64decode(payload).decode("utf-8")
                obj_json = json.loads(obj_file)
                if setting_value is False:
                    if setting_name in obj_json.get(setting_type, {}):
                        del obj_json[setting_type][setting_name]
                else:
                    if setting_name not in obj_json.get(setting_type, {}):
                        obj_json[setting_type][setting_name] = True

                _add_part(request_body, path, _conv_b64(obj_json))
            else:
                _add_part(request_body, path, payload)

        upd = self.update_report(request_body=request_body)
        if upd == 200:
            print(f"{icons.green_dot}")
        else:
            print(f"{icons.red_dot}")

    def __persist_filters(self, value: Optional[bool] = False):
        """
        Don't allow end user to save filters on this file in the Power BI service.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="isPersistentUserStateDisabled",
            setting_value=value,
        )

    def __hide_visual_header(self, value: Optional[bool] = False):
        """
        Hide the visual header in reading view.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="hideVisualContainerHeader",
            setting_value=value,
        )

    def __default_cross_filtering(self, value: Optional[bool] = False):
        """
        Change the default visual interaction from cross highlighting to cross filtering.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="defaultFilterActionIsDataFilter",
            setting_value=value,
        )

    def __modern_visual_header(self, value: Optional[bool] = True):
        """
        Use the modern visual header with updated styling options.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="useStylableVisualContainerHeader",
            setting_value=value,
        )

    def __show_default_summarization_type(self, value: Optional[bool] = True):
        """
        For aggregated fields, always show the default summarization type.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="useDefaultAggregateDisplayName",
            setting_value=value,
        )

    def __modern_visual_tooltips(self, value: Optional[bool] = True):
        """
        Use modern visual tooltips with drill actions and updated styling.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="useEnhancedTooltips",
            setting_value=value,
        )

    def __user_can_change_filter_types(self, value: Optional[bool] = True):
        """
        Allow users to change filter types.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="allowChangeFilterTypes",
            setting_value=value,
        )

    def __disable_search_filter_pane(self, value: Optional[bool] = False):
        """
        Enable search for the filter pane.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="disableFilterPaneSearch",
            setting_value=value,
        )

    def __enable_cross_report_drillthrough(self, value: Optional[bool] = False):
        """
        Allow visuals in this report to use drillthrough targets from other reports.
        """

        self.adjust_settings(
            setting_type="settings",
            setting_name="useCrossReportDrillthrough",
            setting_value=value,
        )

    def __disable_default_cross_highlighting(self, value: Optional[bool] = False):
        """
        Disable cross highlighting/filtering by default.
        """

        self.adjust_settings(
            setting_type="slowDataSourceSettings",
            setting_name="isCrossHighlightingDisabled",
            setting_value=value,
        )

    def __add_slicer_apply_button(self, value: Optional[bool] = False):
        """
        Add an Apply button to each individual slicer (not recommended).
        """

        self.adjust_settings(
            setting_type="slowDataSourceSettings",
            setting_name="isSlicerSelectionsButtonEnabled",
            setting_value=value,
        )

    # def close(self):
    # if not self._readonly and self._report is not None:
    #    print("saving...")

    #    self._report = None
